#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""Diode NetBox Plugin - Diode - Auth."""

import datetime
import json
import logging
import re
import threading
from dataclasses import dataclass
from urllib.parse import urlencode

import requests

from netbox_diode_plugin.plugin_config import (
    get_diode_auth_base_url,
    get_diode_credentials,
    get_diode_max_auth_retries,
)

SCOPE_DIODE_READ = "diode:read"
SCOPE_DIODE_WRITE = "diode:write"

logger = logging.getLogger("netbox.diode_data")

valid_client_id_re = re.compile(r"^[a-zA-Z0-9_-]{1,64}$")

_client = None
_client_lock = threading.Lock()
def get_api_client():
    """Get the client API client."""
    global _client
    global _client_lock

    with _client_lock:
        if _client is None:
            client_id, client_secret = get_diode_credentials()
            if not client_id:
                raise ClientAPIError(
                    "Please update the plugin configuration to access this feature.\nMissing netbox to diode client id.", 500)
            if not client_secret:
                raise ClientAPIError(
                    "Please update the plugin configuration to access this feature.\nMissing netbox to diode client secret.", 500)
            max_auth_retries = get_diode_max_auth_retries()
            _client = ClientAPI(
                base_url=get_diode_auth_base_url(),
                client_id=client_id,
                client_secret=client_secret,
                max_auth_retries=max_auth_retries,
            )
        return _client


class ClientAPIError(Exception):
    """Client API Error."""

    def __init__(self, message: str, status_code: int = 500):
        """Initialize the ClientAPIError."""
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)

    def is_auth_error(self) -> bool:
        """Check if the error is an authentication error."""
        return self.status_code == 401 or self.status_code == 403

class ClientAPI:
    """Manages Diode Clients."""

    def __init__(self, base_url: str, client_id: str, client_secret: str, max_auth_retries: int = 2):
        """Initialize the ClientAPI."""
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret

        self._max_auth_retries = max_auth_retries
        self._client_auth_token = None
        self._client_auth_token_lock = threading.Lock()

    def create_client(self, name: str, scope: str) -> dict:
        """Create a client."""
        for attempt in range(self._max_auth_retries):
            token = None
            try:
                token = self._get_token()
                url = self.base_url + "/clients"
                headers = {"Authorization": f"Bearer {token}"}
                data = {
                    "client_name": name,
                    "scope": scope,
                }
                response = requests.post(url, json=data, headers=headers)
                if response.status_code != 201:
                    raise ClientAPIError("Failed to create client", response.status_code)
                return response.json()
            except ClientAPIError as e:
                if e.is_auth_error() and attempt < self._max_auth_retries - 1:
                    logger.info(f"Retrying create_client due to unauthenticated error, attempt {attempt + 1}")
                    self._mark_client_auth_token_invalid(token)
                    continue
                raise
        raise ClientAPIError("Failed to create client: unexpected state", 500)

    def get_client(self, client_id: str) -> dict:
        """Get a client."""
        if not valid_client_id_re.match(client_id):
            raise ValueError(f"Invalid client ID: {client_id}")

        for attempt in range(self._max_auth_retries):
            token = None
            try:
                token = self._get_token()
                url = self.base_url + f"/clients/{client_id}"
                headers = {"Authorization": f"Bearer {token}"}
                response = requests.get(url, headers=headers)
                if response.status_code == 401 or response.status_code == 403:
                    raise ClientAPIError(f"Failed to get client {client_id}", response.status_code)
                if response.status_code != 200:
                    raise ClientAPIError(f"Failed to get client {client_id}", response.status_code)
                return response.json()
            except ClientAPIError as e:
                if e.is_auth_error() and attempt < self._max_auth_retries - 1:
                    logger.info(f"Retrying delete_client due to unauthenticated error, attempt {attempt + 1}")
                    self._mark_client_auth_token_invalid(token)
                    continue
                raise
        raise ClientAPIError(f"Failed to get client {client_id}: unexpected state")

    def delete_client(self, client_id: str) -> None:
        """Delete a client."""
        if not valid_client_id_re.match(client_id):
            raise ValueError(f"Invalid client ID: {client_id}")

        for attempt in range(self._max_auth_retries):
            token = None
            try:
                token = self._get_token()
                url = self.base_url + f"/clients/{client_id}"
                headers = {"Authorization": f"Bearer {token}"}
                response = requests.delete(url, headers=headers)
                if response.status_code != 204:
                    raise ClientAPIError(f"Failed to delete client {client_id}", response.status_code)
                return
            except ClientAPIError as e:
                if e.is_auth_error() and attempt < self._max_auth_retries - 1:
                    logger.info(f"Retrying delete_client due to unauthenticated error, attempt {attempt + 1}")
                    self._mark_client_auth_token_invalid(token)
                    continue
                raise
        raise ClientAPIError(f"Failed to delete client {client_id}: unexpected state")


    def list_clients(self, page_token: str | None = None, page_size: int | None = None) -> list[dict]:
        """List all clients."""
        for attempt in range(self._max_auth_retries):
            token = None
            try:
                token = self._get_token()
                url = self.base_url + "/clients"
                headers = {"Authorization": f"Bearer {token}"}
                params = {}
                if page_token:
                    params["page_token"] = page_token
                if page_size:
                    params["page_size"] = page_size
                response = requests.get(url, headers=headers, params=params)
                if response.status_code != 200:
                    raise ClientAPIError("Failed to get clients", response.status_code)
                return response.json()
            except ClientAPIError as e:
                if e.is_auth_error() and attempt < self._max_auth_retries - 1:
                    logger.info(f"Retrying list_clients due to unauthenticated error, attempt {attempt + 1}")
                    self._mark_client_auth_token_invalid(token)
                    continue
                raise
        raise ClientAPIError("Failed to list clients: unexpected state")


    def _get_token(self) -> str:
        """Get a token for the Diode Auth Service."""
        with self._client_auth_token_lock:
            if self._client_auth_token:
                return self._client_auth_token
            self._client_auth_token = self._authenticate()
            return self._client_auth_token

    def _mark_client_auth_token_invalid(self, token: str):
        """Mark a client auth token as invalid."""
        with self._client_auth_token_lock:
            self._client_auth_token = None

    def _authenticate(self) -> str:
        """Get a new access token for the Diode Auth Service."""
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = urlencode(
            {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "scope": f"{SCOPE_DIODE_READ} {SCOPE_DIODE_WRITE}",
            }
        )
        url = self.base_url + "/token"
        try:
            response = requests.post(url, data=data, headers=headers)
        except Exception as e:
            raise ClientAPIError(f"Failed to obtain access token: {e}", 401) from e
        if response.status_code != 200:
            raise ClientAPIError(f"Failed to obtain access token: {response.reason}", 401)

        try:
            token_info = response.json()
        except Exception as e:
            raise ClientAPIError(f"Failed to parse access token response: {e}", 401) from e

        access_token = token_info.get("access_token")
        if not access_token:
            raise ClientAPIError(f"Failed to obtain access token for client {self._client_id}", 401)

        return access_token




#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin."""

from netbox.plugins import PluginConfig

from .version import version_semver


class NetBoxDiodePluginConfig(PluginConfig):
    """NetBox Diode plugin configuration."""

    name = "netbox_diode_plugin"
    verbose_name = "NetBox Labs, Diode Plugin"
    description = "Diode plugin for NetBox."
    version = version_semver()
    base_url = "diode"
    min_version = "4.4.10"
    max_version = "4.5.99"
    default_settings = {
        # Default Diode gRPC target for communication with Diode server
        "diode_target": "grpc://localhost:8080/diode",

        # Default username associated with changes applied via plugin
        "diode_username": "diode",

        # client_id and client_secret for communication with Diode server.
        # By default, the secret is read from a file /run/secrets/netbox_to_diode
        # but may be specified directly as a string in netbox_to_diode_client_secret
        "netbox_to_diode_client_id": "netbox-to-diode",
        "netbox_to_diode_client_secret": None,
        "secrets_path": "/run/secrets/",
        "netbox_to_diode_client_secret_name": "netbox_to_diode",
        "diode_max_auth_retries": 3,

        # List of audiences to require for the diode-to-netbox token.
        # If empty, no audience is required.
        "required_token_audience": [],
    }


config = NetBoxDiodePluginConfig



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Version stamp."""

# These properties are injected at build time by the build process.

__commit_hash__ = "26476e8"
__track__ = "release"
__version__ = "1.7.0"


def version_display():
    """Display the version, track and hash together."""
    return f"v{__version__}-{__track__}-{__commit_hash__}"


def version_semver():
    """Semantic version."""
    return __version__



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode Netbox Plugin - URLs."""

from django.urls import path

from . import views

urlpatterns = (
    path("settings/", views.SettingsView.as_view(), name="settings"),
    path("settings/edit/", views.SettingsEditView.as_view(), name="settings_edit"),
    path("credentials/", views.ClientCredentialListView.as_view(), name="client_credential_list"),
    path("credentials/add/", views.ClientCredentialAddView.as_view(), name="client_credential_add"),
    path("credentials/secret/", views.ClientCredentialSecretView.as_view(), name="client_credential_secret"),
    path("credentials/delete/<str:client_credential_id>/", views.ClientCredentialDeleteView.as_view(), name="client_credential_delete"),
)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Navigation."""

from django.utils.translation import gettext as _
from netbox.plugins import PluginMenu, PluginMenuItem

_diode_menu_items = (
    PluginMenuItem(
        link="plugins:netbox_diode_plugin:settings",
        link_text=_("Settings"),
        permissions=("netbox_diode_plugin.view_setting",),
    ),
    PluginMenuItem(
        link="plugins:netbox_diode_plugin:client_credential_list",
        link_text=_("Client Credentials"),
        permissions=("netbox_diode_plugin.view_clientcredentials",),
    ),
)

menu = PluginMenu(
    label="Diode",
    groups=(
        (_("Diode"), _diode_menu_items),
    ),
    icon_class="mdi mdi-upload",
)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API Authentication."""

import hashlib
import logging
from types import SimpleNamespace

import requests
from django.core.cache import cache
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed

from netbox_diode_plugin.plugin_config import (
    get_diode_auth_introspect_url,
    get_diode_user,
    get_required_token_audience,
)

logger = logging.getLogger("netbox.diode_data")


class DiodeOAuth2Authentication(BaseAuthentication):
    """Diode OAuth2 Client Credentials Authentication."""

    def authenticate(self, request):
        """Authenticate the request and return the user info."""
        auth_header = request.headers.get("Authorization", "")
        if not auth_header.startswith("Bearer "):
            return None

        token = auth_header[7:].strip()

        diode_user = self._introspect_token(token)
        if not diode_user:
            raise AuthenticationFailed("Invalid token")

        request.user = diode_user.user
        request.token_scopes = diode_user.token_scopes
        request.token_data = diode_user.token_data

        return (diode_user.user, None)

    def _introspect_token(self, token: str):
        """Introspect the token and return the client info."""
        hash_token = hashlib.sha256(token.encode()).hexdigest()
        cache_key = f"diode:oauth2:introspect:{hash_token}"
        cached_user = cache.get(cache_key)
        if cached_user:
            return cached_user

        introspect_url = get_diode_auth_introspect_url()

        if not introspect_url:
            logger.error("Diode Auth introspect URL is not configured")
            return None

        try:
            response = requests.post(
                introspect_url, headers={"Authorization": f"Bearer {token}"}, timeout=5
            )
            response.raise_for_status()
            data = response.json()
        except Exception as e:
            logger.error(f"Diode Auth token introspection failed: {e}")
            return None

        if data.get("active"):
            # if the plugin is configured to require specific token audience(s),
            # reject the token if any are missing.
            required_audience = get_required_token_audience()
            if required_audience:
                token_audience = set(data.get("aud", []))
                missing_audience = set(required_audience) - token_audience
                if missing_audience:
                    logger.error(f"Token audience(s) {missing_audience} not found in {token_audience}")
                    return None

            diode_user = SimpleNamespace(
                user=get_diode_user(),
                token_scopes=data.get("scope", "").split(),
                token_data=data,
            )

            expires_in = (
                data.get("exp") - data.get("iat")
                if "exp" in data and "iat" in data
                else 300
            )
            cache.set(cache_key, diode_user, timeout=expires_in)
            return diode_user

        return None



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Database migrations."""

from django.db import migrations, models


class Migration(migrations.Migration):
    """Clean up Setting model by removing unused fields."""

    dependencies = [
        ("netbox_diode_plugin", "0006_clientcredentials_alter_setting_diode_target"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="setting",
            name="custom_field_data",
        ),
        migrations.RemoveField(
            model_name='setting',
            name='created',
        ),
        migrations.RemoveField(
            model_name='setting',
            name='last_updated',
        ),
    ]



#!/usr/bin/env python
"""Django management command to generate markdown documentation for NetBox Diode Plugin matching criteria."""

from dataclasses import dataclass
from typing import Optional

from django.core.management.base import BaseCommand

from netbox_diode_plugin.api.differ import extract_supported_models
from netbox_diode_plugin.api.matcher import _LOGICAL_MATCHERS, get_model_matchers


@dataclass
class MatcherInfo:
    """Information about a matcher for documentation."""

    name: str
    fields: list[str] | None = None
    condition: str | None = None
    description: str | None = None
    matcher_type: str = "ObjectMatchCriteria"
    version_constraints: str | None = None
    matcher_source: str = "logical"  # "logical" or "builtin"


class Command(BaseCommand):
    """Django management command to generate markdown documentation for NetBox Diode Plugin matching criteria."""

    help = "Generate markdown documentation for NetBox Diode Plugin matching criteria"

    def extract_condition_description(self, condition) -> str:
        """Extract a human-readable description of a Q condition."""
        if condition is None:
            return "None"

        # Handle simple conditions
        if hasattr(condition, 'children'):
            conditions = []
            for child in condition.children:
                if isinstance(child, tuple):
                    field, value = child
                    if field.endswith('__isnull'):
                        field_name = field[:-8]
                        if value:
                            conditions.append(f"{field_name} is NULL")
                        else:
                            conditions.append(f"{field_name} is NOT NULL")
                    else:
                        conditions.append(f"{field} = {value}")
                else:
                    conditions.append(str(child))

            connector = " AND " if condition.connector == "AND" else " OR "
            return connector.join(conditions)

        return str(condition)

    def get_matcher_description(self, matcher) -> str:  # noqa: C901
        """Generate a human-readable description of what the matcher does."""
        # Handle IP Network matchers
        if hasattr(matcher, 'ip_fields') and matcher.ip_fields and hasattr(matcher, 'vrf_field') and matcher.vrf_field:
            ip_fields_str = ", ".join(matcher.ip_fields)
            if matcher.name.startswith('logical_ip_address_global_no_vrf'):
                return f"Matches IP address {ip_fields_str} in global namespace (no VRF)"
            if matcher.name.startswith('logical_ip_address_within_vrf'):
                return f"Matches IP address {ip_fields_str} within VRF"
            if matcher.name.startswith('logical_ip_range'):
                return f"Matches IP range {ip_fields_str} within VRF context"

        # Handle CustomFieldMatcher
        if hasattr(matcher, 'custom_field') and matcher.custom_field:
            return f"Matches on unique custom field: {matcher.custom_field}"

        # Handle AutoSlugMatcher
        if hasattr(matcher, 'slug_field') and matcher.slug_field:
            return f"Matches on auto-generated slug field: {matcher.slug_field}"

        # Handle builtin unique field matchers
        if matcher.name.startswith('unique_') and hasattr(matcher, 'fields') and matcher.fields:
            field_name = matcher.fields[0] if len(matcher.fields) == 1 else ", ".join(matcher.fields)
            if matcher.name.startswith('unique_'):
                return f"Matches on unique field(s): {field_name}"

        # Handle builtin UniqueConstraint matchers
        if hasattr(matcher, 'fields') and matcher.fields and not matcher.name.startswith('logical_'):
            fields_str = ", ".join(matcher.fields)
            if hasattr(matcher, 'condition') and matcher.condition:
                condition_desc = self.extract_condition_description(matcher.condition)
                return f"Matches on unique constraint fields: {fields_str} where {condition_desc}"
            return f"Matches on unique constraint fields: {fields_str}"

        # Standard field-based matcher
        if hasattr(matcher, 'fields') and matcher.fields:
            fields_str = ", ".join(matcher.fields)
            if hasattr(matcher, 'condition') and matcher.condition:
                condition_desc = self.extract_condition_description(matcher.condition)
                return f"Matches on fields: {fields_str} where {condition_desc}"
            return f"Matches on fields: {fields_str}"

        return "Custom matcher"

    def get_version_constraints(self, matcher) -> str | None:
        """Get version constraints as a string."""
        constraints = []
        if hasattr(matcher, 'min_version') and matcher.min_version:
            constraints.append(f"≥{matcher.min_version}")
        if hasattr(matcher, 'max_version') and matcher.max_version:
            constraints.append(f"≤{matcher.max_version}")

        return " ".join(constraints) if constraints else None

    def analyze_logical_matchers(self) -> dict[str, list[MatcherInfo]]:
        """Analyze the logical matchers and extract documentation information."""
        documentation = {}

        for object_type, matcher_factory in _LOGICAL_MATCHERS.items():
            matchers = matcher_factory()
            matcher_infos = []

            for matcher in matchers:
                info = MatcherInfo(
                    name=matcher.name,
                    fields=list(matcher.fields) if hasattr(matcher, 'fields') and matcher.fields else None,
                    condition=self.extract_condition_description(matcher.condition) if hasattr(matcher, 'condition') else None,
                    description=self.get_matcher_description(matcher),
                    matcher_type=matcher.__class__.__name__,
                    version_constraints=self.get_version_constraints(matcher),
                    matcher_source="logical"
                )
                matcher_infos.append(info)

            documentation[object_type] = matcher_infos

        return documentation

    def analyze_builtin_matchers(self) -> dict[str, list[MatcherInfo]]:
        """Analyze the builtin matchers and extract documentation information."""
        documentation = {}

        for object_type, model_info in extract_supported_models().items():
            model_class = model_info["model"]
            matchers = get_model_matchers(model_class)
            matcher_infos = []

            for matcher in matchers:
                # Skip logical matchers as they're already handled
                if matcher.name.startswith('logical_'):
                    continue

                # Extract fields for builtin matchers
                fields = None
                if hasattr(matcher, 'fields') and matcher.fields:
                    fields = list(matcher.fields)
                elif hasattr(matcher, 'custom_field'):
                    fields = [f"custom_fields.{matcher.custom_field}"]
                elif hasattr(matcher, 'slug_field'):
                    fields = [matcher.slug_field]

                info = MatcherInfo(
                    name=matcher.name,
                    fields=fields,
                    condition=self.extract_condition_description(matcher.condition) if hasattr(matcher, 'condition') else None,
                    description=self.get_matcher_description(matcher),
                    matcher_type=matcher.__class__.__name__,
                    version_constraints=self.get_version_constraints(matcher),
                    matcher_source="builtin"
                )
                matcher_infos.append(info)

            if matcher_infos:  # Only add if there are builtin matchers
                documentation[object_type] = matcher_infos

        return documentation

    def combine_matchers(
        self,
        logical_docs: dict[str, list[MatcherInfo]],
        builtin_docs: dict[str, list[MatcherInfo]],
    ) -> dict[str, list[MatcherInfo]]:
        """Combine logical and builtin matchers into a single documentation structure."""
        combined = {}

        # Get all object types
        all_object_types = set(logical_docs.keys()) | set(builtin_docs.keys())

        for object_type in all_object_types:
            matchers = []

            # Add logical matchers
            if object_type in logical_docs:
                matchers.extend(logical_docs[object_type])

            # Add builtin matchers
            if object_type in builtin_docs:
                matchers.extend(builtin_docs[object_type])

            if matchers:
                combined[object_type] = matchers

        return combined

    def generate_markdown_table(self, docs: dict[str, list[MatcherInfo]]) -> str:
        """Generate a markdown table from the documentation."""
        markdown = []
        markdown.append("# NetBox Diode Plugin - Object Matching Criteria")
        markdown.append("")
        markdown.append(
            "This document describes how the Diode NetBox Plugin matches existing objects when applying changes. "
            "The matchers will be applied in the order of their precedence, unttil one of them matches."
        )
        markdown.append("")
        markdown.append("## Matcher Types")
        markdown.append("")
        markdown.append("- **Logical Matchers**: Custom matching criteria that represent likely user intent")
        markdown.append(
            "- **Builtin Matchers**: Automatically generated from NetBox model constraints "
            "(unique fields, unique constraints, custom fields, auto-slugs)"
        )
        markdown.append("")

        # Sort object types for consistent output
        sorted_object_types = sorted(docs.keys())

        for object_type in sorted_object_types:
            matchers = docs[object_type]

            markdown.append(f"## {object_type}")
            markdown.append("")

            if not matchers:
                markdown.append("No specific matching criteria defined.")
                markdown.append("")
                continue

            # Create table header
            markdown.append("| Matcher Name | Order of Precedence | Type | Fields | Condition | Description | Version Constraints |")
            markdown.append("|--------------|---------------------|------|--------|-----------|-------------|---------------------|")

            for precedence, matcher in enumerate(matchers, start=1):
                # Escape pipe characters in table cells
                name = matcher.name.replace("|", "\\|") if matcher.name else "N/A"
                matcher_type = matcher.matcher_source.replace("|", "\\|")
                fields_str = ", ".join(matcher.fields).replace("|", "\\|") if matcher.fields else ""
                condition_str = matcher.condition.replace("|", "\\|") if matcher.condition and matcher.condition != "None" else "N/A"
                description = matcher.description.replace("|", "\\|") if matcher.description else "N/A"
                version_str = matcher.version_constraints.replace("|", "\\|") if matcher.version_constraints else "All versions"

                markdown.append(
                    f"| {name} | {precedence} | {matcher_type} | {fields_str} | {condition_str} | {description} | {version_str} |"
                )

            markdown.append("")

        return "\n".join(markdown)

    def handle(self, *args, **options):
        """Handle the command execution."""
        self.stdout.write("Analyzing logical matching criteria...")
        logical_docs = self.analyze_logical_matchers()

        self.stdout.write("Analyzing builtin matching criteria...")
        builtin_docs = self.analyze_builtin_matchers()

        self.stdout.write("Combining matchers...")
        combined_docs = self.combine_matchers(logical_docs, builtin_docs)

        self.stdout.write("Generating markdown documentation...")
        markdown_content = self.generate_markdown_table(combined_docs)
        self.stdout.write(markdown_content)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Views."""
import logging
from collections import defaultdict

from django.conf import settings as netbox_settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import gettext as _
from django.views.generic import View
from netbox.plugins import get_plugin_config
from netbox.views import generic
from users.models import ObjectPermission
from utilities.forms import ConfirmationForm
from utilities.htmx import htmx_partial
from utilities.permissions import get_permission_for_model, permission_is_exempt
from utilities.views import register_model_view

from netbox_diode_plugin.client import create_client, delete_client, get_client, list_clients
from netbox_diode_plugin.forms import ClientCredentialForm, SettingsForm
from netbox_diode_plugin.models import ClientCredentials, Setting
from netbox_diode_plugin.tables import ClientCredentialsTable

User = get_user_model()


logger = logging.getLogger(__name__)

def redirect_to_login(request):
    """Redirect to login view."""
    redirect_url = netbox_settings.LOGIN_URL
    target = request.path

    if target and url_has_allowed_host_and_scheme(target, allowed_hosts=None):
        redirect_url = f"{netbox_settings.LOGIN_URL}?next={target}"

    return HttpResponseRedirect(redirect_url)


class BaseDiodeView(View):
    """
    Base view class for Diode plugin views.

    Provides authentication and permission checking functionality for views
    that need to interact with the Diode API. Includes methods for:
    - Object permission filtering and retrieval
    - Permission checking for authenticated users
    - Authentication validation for requests
    """

    def get_permission_filter(self, user_obj):
        """Return the permission filter for the user."""
        return Q(users=user_obj) | Q(groups__user=user_obj)

    def get_object_permissions(self, user_obj):
        """Return all permissions granted to the user by an ObjectPermission."""
        # Initialize a dictionary mapping permission names to sets of constraints
        perms = defaultdict(list)

        # Collect any configured default permissions
        for perm_name, constraints in netbox_settings.DEFAULT_PERMISSIONS.items():
            constraints = constraints or ()
            if type(constraints) not in (list, tuple):
                raise ImproperlyConfigured(
                    f"Constraints for default permission {perm_name} must be defined as a list or tuple."
                )
            perms[perm_name].extend(constraints)

        # Retrieve all assigned and enabled ObjectPermissions
        object_permissions = ObjectPermission.objects.filter(
            self.get_permission_filter(user_obj),
            enabled=True
        ).order_by('id').distinct('id').prefetch_related('object_types')

        # Create a dictionary mapping permissions to their constraints
        for obj_perm in object_permissions:
            for object_type in obj_perm.object_types.all():
                for action in obj_perm.actions:
                    perm_name = f"{object_type.app_label}.{action}_{object_type.model}"
                    perms[perm_name].extend(obj_perm.list_constraints())

        return perms

    def get_all_permissions(self, user_obj, obj=None):
        """Get all permissions for the user."""
        if not user_obj.is_active or user_obj.is_anonymous:
            return {}
        if not hasattr(user_obj, '_object_perm_cache'):
            user_obj._object_perm_cache = self.get_object_permissions(user_obj)
        return user_obj._object_perm_cache

    def has_perm(self, user_obj, perm):
        """Check if the user has the required permission."""
        # Superusers implicitly have all permissions
        if not user_obj.is_authenticated:
            return False

        if user_obj.is_active and user_obj.is_superuser:
            return True

        # Permission is exempt from enforcement (i.e. listed in EXEMPT_VIEW_PERMISSIONS)
        if permission_is_exempt(perm):
            return True

        # Handle inactive/anonymous users
        if not user_obj.is_active or user_obj.is_anonymous:
            return False

        object_permissions = self.get_all_permissions(user_obj)

        # If no applicable ObjectPermissions have been created for this user/permission, deny permission
        if perm not in object_permissions:
            return False

        return True

    def check_authentication(self, request):
        """Check if the user has the required permission."""
        if not request.user.is_authenticated:
            return redirect_to_login(request)

        if not self.has_perm(request.user, self.get_required_permission()):
            return redirect(
                reverse("home",)
            )
        return None

class SettingsView(BaseDiodeView):
    """Settings view."""

    def get_required_permission(self):
        """Return the permission required to view Diode plugin settings."""
        return "netbox_diode_plugin.view_setting"

    def get(self, request):
        """Render settings template."""
        if ret := self.check_authentication(request):
            return ret

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )

        try:
            settings = Setting.objects.get()
        except Setting.DoesNotExist:
            default_diode_target = get_plugin_config(
                "netbox_diode_plugin", "diode_target"
            )
            settings = Setting.objects.create(
                diode_target=diode_target_override or default_diode_target
            )

        diode_target = diode_target_override or settings.diode_target

        # Check if branching plugin is available
        from django.conf import settings as django_settings
        has_branching_plugin = "netbox_branching" in django_settings.PLUGINS

        context = {
            "diode_target": diode_target,
            "is_diode_target_overridden": diode_target_override is not None,
            "branch": settings.branch if has_branching_plugin else None,
            "has_branching_plugin": has_branching_plugin,
        }

        return render(request, "diode/settings.html", context)


@register_model_view(Setting, "edit")
class SettingsEditView(BaseDiodeView,generic.ObjectEditView):
    """Settings edit view."""

    queryset = Setting.objects
    form = SettingsForm
    template_name = "diode/settings_edit.html"
    default_return_url = "plugins:netbox_diode_plugin:settings"

    def get_required_permission(self):
        """Return the permission required to view Diode plugin settings."""
        return "netbox_diode_plugin.change_setting"

    def get(self, request, *args, **kwargs):
        """GET request handler."""
        if ret := self.check_authentication(request):
            return ret

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )
        if diode_target_override:
            messages.info(
                request,
                "The Diode target field is disabled because it is overridden in the plugin configuration.",
            )

        settings = Setting.objects.get()
        kwargs["pk"] = settings.pk

        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """POST request handler."""
        if ret := self.check_authentication(request):
            return ret

        settings = Setting.objects.get()
        kwargs["pk"] = settings.pk

        return super().post(request, *args, **kwargs)


class GetReturnURLMixin:
    """Get return URL mixin."""

    def get_return_url(self, request):
        """Get return URL."""
        # First, see if `return_url` was specified as a query parameter or form data. Use this URL only if it's
        # considered safe.
        return_url = request.GET.get("return_url") or request.POST.get("return_url")
        if return_url and url_has_allowed_host_and_scheme(
            return_url, allowed_hosts=None
        ):
            return return_url

        return None


class ClientCredentialListView(BaseDiodeView):
    """Client credential list view."""

    table = ClientCredentialsTable
    template_name = "diode/client_credential_list.html"
    model = ClientCredentials

    def get_required_permission(self):
        """Return the permission required to view client credentials list."""
        return "netbox_diode_plugin.view_clientcredentials"

    def get_table_data(self, request):
        """Get table data."""
        try:
            data = list_clients(request)
            total = len(data)
        except Exception as e:
            logger.debug(f"Error loading client credentials error: {str(e)}")
            messages.error(self.request, str(e))
            data = []
            total = 0

        return total, data

    def get(self, request):
        """GET request handler."""
        if ret := self.check_authentication(request):
            return ret

        total, data = self.get_table_data(request)
        table = self.table(data=data)  # Pass the data to the table

        # If this is an HTMX request, return only the rendered table HTML
        if htmx_partial(request):
            if request.GET.get("embedded", False):
                table.embedded = True
                # Hide selection checkboxes
                if "pk" in table.base_columns:
                    table.columns.hide("pk")
            return render(
                request,
                "htmx/table.html",
                {
                    "model": ClientCredentials,
                    "table": table,
                    "total_count": len(data),
                },
            )

        context = {
            "model": ClientCredentials,
            "table": table,
            "total_count": len(data),
        }

        return render(request, self.template_name, context)


class ClientCredentialDeleteView(GetReturnURLMixin, BaseDiodeView):
    """Client credential delete view."""

    template_name = "diode/client_credential_delete.html"
    default_return_url = "plugins:netbox_diode_plugin:client_credential_list"

    def get_required_permission(self):
        """Return the permission required to delete client credentials."""
        return "netbox_diode_plugin.delete_clientcredentials"

    def get(self, request, client_credential_id):
        """GET request handler."""
        if ret := self.check_authentication(request):
            return ret

        data = get_client(request, client_credential_id)

        return render(
            request,
            self.template_name,
            {
                "object": data,
                "object_type": "Client Credential",
                "return_url": self.get_return_url(request) or reverse(self.default_return_url),
            },
        )

    def post(self, request, client_credential_id):
        """POST request handler."""
        sanitized_client_credential_id = client_credential_id.replace('\n', '').replace('\r', '')
        logger.info(f"Deleting client {sanitized_client_credential_id}")
        if ret := self.check_authentication(request):
            return ret

        form = ConfirmationForm(request.POST)
        if form.is_valid():
            try:
                delete_client(request, client_credential_id)
                messages.success(request, _("Client deleted successfully"))
            except Exception as e:
                logger.error(
                    f"Error deleting client: {sanitized_client_credential_id} error: {str(e)}"
                )
                messages.error(request, str(e))

        return redirect(
            reverse(
                "plugins:netbox_diode_plugin:client_credential_list",
            )
        )


class ClientCredentialAddView(GetReturnURLMixin, BaseDiodeView):
    """View for adding client credentials."""

    template_name = "diode/client_credential_add.html"
    form_class = ClientCredentialForm
    default_return_url = "plugins:netbox_diode_plugin:client_credential_list"

    def get_required_permission(self):
        """Return the permission required to add new client credentials."""
        return "netbox_diode_plugin.add_clientcredentials"

    def get(self, request):
        """GET request handler."""
        if ret := self.check_authentication(request):
            return ret

        form = self.form_class()
        return render(
            request,
            self.template_name,
            {
                "form": form,
                "return_url": self.get_return_url(request) or reverse(self.default_return_url),
            },
        )

    def post(self, request):
        """POST request handler."""
        if ret := self.check_authentication(request):
            return ret

        form = self.form_class(request.POST)
        if form.is_valid():
            try:
                response = create_client(request, form.cleaned_data["client_name"], "diode:ingest")
                # Store the client credentials in session
                request.session['client_secret'] = response.get('client_secret')
                request.session['client_name'] = form.cleaned_data["client_name"]
                request.session['client_id'] = response.get('client_id')
                return redirect(
                    reverse(
                        "plugins:netbox_diode_plugin:client_credential_secret",
                    )
                )
            except Exception as e:
                logger.error(f"Error creating client: {str(e)}")
                messages.error(request, str(e))

        return render(
            request,
            self.template_name,
            {
                "form": form,
                "return_url": self.get_return_url(request) or reverse(self.default_return_url),
            },
        )


class ClientCredentialSecretView(BaseDiodeView):
    """View for displaying client secret."""

    template_name = "diode/client_credential_secret.html"

    def get_required_permission(self):
        """Return the permission required to view client credential secrets."""
        return "netbox_diode_plugin.view_clientcredentials"

    def get(self, request):
        """Get request handler."""
        if ret := self.check_authentication(request):
            return ret

        # Get the client secret from session
        client_secret = request.session.get('client_secret')
        client_name = request.session.get('client_name')
        client_id = request.session.get('client_id')

        if not client_secret:
            messages.error(request, _("No client secret found. Please create a new client."))
            return redirect(
                reverse(
                    "plugins:netbox_diode_plugin:client_credential_list",
                )
            )

        # Clear the session data after retrieving it
        request.session.pop('client_secret', None)
        request.session.pop('client_name', None)
        request.session.pop('client_id', None)

        return render(
            request,
            self.template_name,
            {
                "object": {
                    "client_name": client_name,
                    "client_id": client_id,
                    "client_secret": client_secret,
                }
            },
        )
